from __future__ import annotations

import argparse
import importlib
from importlib import metadata
import os
from pathlib import Path
import subprocess
import sys

from . import resolve_cf_pipeline_v2_executable

try:
    from cf_cli.registry import register_command
except ImportError:  # pragma: no cover - optional dependency
    register_command = None


DIRECT_ENGINE_ALLOW_ENV = "CF_ALLOW_DIRECT_ENGINE"
UNIFIED_ENGINE_REGISTER_ENV = "CF_ENABLE_UNIFIED_ENGINE_COMMAND"


def _env_enabled(name: str) -> bool:
    raw = os.environ.get(name, "")
    return raw.strip().lower() in {"1", "true", "yes", "on"}


def _direct_engine_allowed() -> bool:
    return _env_enabled(DIRECT_ENGINE_ALLOW_ENV)


def _policy_error_message() -> str:
    return (
        "Direct engine invocation is blocked by policy. "
        "Use 'python -m cf_ontology state activate -> state emit-event -> state run-loop'. "
        f"For explicit override only, set {DIRECT_ENGINE_ALLOW_ENV}=1."
    )


def _find_default_exe() -> str:
    return resolve_cf_pipeline_v2_executable()


def _add_engine_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument("--pipeline", required=True, help="Path to pipeline JSON-LD")
    parser.add_argument("--steps", action="append", help="Path to steps JSON-LD (repeatable)")
    parser.add_argument(
        "--plugins",
        action="append",
        help="Directory containing step plugins (optional)",
    )
    parser.add_argument(
        "--engine-threads",
        "--threads",
        dest="engine_threads",
        type=int,
        default=0,
        help="Engine worker threads (0=auto)",
    )
    parser.add_argument(
        "--max-total-threads",
        type=int,
        default=None,
        help="Global thread budget cap (default=engine threads)",
    )
    parser.add_argument(
        "--default-step-threads",
        type=int,
        default=None,
        help="Default per-step internal thread budget",
    )
    parser.add_argument(
        "--respect-step-parallelism",
        nargs="?",
        const="true",
        default=None,
        choices=["true", "false"],
        help="Honor step parallelism metadata (default=true)",
    )
    parser.add_argument(
        "--deterministic",
        nargs="?",
        const="true",
        default=None,
        choices=["true", "false"],
        help="Enable deterministic mode (conservative threading)",
    )
    parser.add_argument(
        "--interval",
        type=float,
        default=0.0,
        help="Seconds between repeated runs (0=single run)",
    )
    parser.add_argument(
        "--iterations",
        type=int,
        default=1,
        help="Number of iterations when interval > 0",
    )
    parser.add_argument(
        "--duration",
        type=float,
        default=0.0,
        help="Total duration in seconds (overrides iterations)",
    )
    parser.add_argument("--debug", action="store_true", help="Enable debug checklist output")
    parser.add_argument(
        "--exe",
        default=_find_default_exe(),
        help="Path to cf_pipeline_v2 executable",
    )


def _discover_plugin_dirs() -> list[str]:
    plugin_dirs: list[str] = []
    seen_pkgs: set[str] = set()
    seen_dirs: set[str] = set()
    try:
        entry_points = metadata.entry_points(group="cogniflow.steps")
    except TypeError:
        entry_points = metadata.entry_points().get("cogniflow.steps", [])
    for ep in entry_points:
        module_name = ep.value.split(":")[0]
        pkg_name = module_name.split(".")[0]
        if pkg_name in seen_pkgs:
            continue
        seen_pkgs.add(pkg_name)
        try:
            module = importlib.import_module(pkg_name)
        except Exception:
            continue
        module_file = getattr(module, "__file__", None)
        if not module_file:
            continue
        pkg_path = Path(module_file).resolve().parent
        bin_dir = pkg_path / "bin"
        if bin_dir.exists():
            resolved = str(bin_dir.resolve())
            if resolved not in seen_dirs:
                seen_dirs.add(resolved)
                plugin_dirs.append(resolved)
    return plugin_dirs


def _run_from_args(args: argparse.Namespace) -> int:
    if not _direct_engine_allowed():
        print(_policy_error_message(), file=sys.stderr)
        return 2

    plugin_dirs = args.plugins or _discover_plugin_dirs()
    if not plugin_dirs:
        print(
            "No plugin directories found. Proceeding without --plugins; "
            "the pipeline header must provide plugin paths.",
            file=sys.stderr,
        )

    cmd = [args.exe, "--pipeline", args.pipeline]
    if args.steps:
        for step_path in args.steps:
            cmd.extend(["--steps", step_path])
    for plugin_dir in plugin_dirs:
        cmd.extend(["--plugins", plugin_dir])
    if args.engine_threads:
        cmd.extend(["--engine-threads", str(args.engine_threads)])
    if args.max_total_threads is not None:
        cmd.extend(["--max-total-threads", str(args.max_total_threads)])
    if args.default_step_threads is not None:
        cmd.extend(["--default-step-threads", str(args.default_step_threads)])
    if args.respect_step_parallelism is not None:
        cmd.extend(["--respect-step-parallelism", str(args.respect_step_parallelism)])
    if args.deterministic is not None:
        cmd.extend(["--deterministic", str(args.deterministic)])
    if args.interval:
        cmd.extend(["--interval", str(args.interval)])
    if args.interval and args.iterations:
        cmd.extend(["--iterations", str(args.iterations)])
    if args.duration:
        cmd.extend(["--duration", str(args.duration)])
    if args.debug:
        cmd.append("--debug")

    try:
        proc = subprocess.run(cmd, check=False)
    except FileNotFoundError:
        print(f"Executable not found: {args.exe}", file=sys.stderr)
        return 2
    return proc.returncode


def _register_cf_cli() -> None:
    if register_command is None:
        return
    if not _env_enabled(UNIFIED_ENGINE_REGISTER_ENV):
        return

    def _add_parser(subparsers: argparse._SubParsersAction) -> argparse.ArgumentParser:
        parser = subparsers.add_parser("engine", help="Run the pipeline engine.")
        _add_engine_args(parser)
        parser.set_defaults(_cf_handler=_run_from_args)
        return parser

    register_command(
        "pipeline",
        "engine",
        _add_parser,
        group_help="Pipeline tools",
        command_help="Run the pipeline engine",
    )


def _register_optional_pipeline_commands() -> None:
    if register_command is None:
        return
    for module_name in (
        "cf_pipeline_cert.cli",
        "cf_pipeline_report.cli",
        "cf_pipeline_report.cli_plugin",
    ):
        try:
            importlib.import_module(module_name)
        except ImportError:
            continue


def main() -> int:
    parser = argparse.ArgumentParser(description="Run a pipeline using the C++ v2 backend.")
    _add_engine_args(parser)
    args = parser.parse_args()
    return _run_from_args(args)


if __name__ == "__main__":
    raise SystemExit(main())


_register_cf_cli()
_register_optional_pipeline_commands()



from __future__ import annotations

import argparse
import os
import statistics
import subprocess
import sys
import time
from pathlib import Path


def _default_exe() -> str:
    engine_dir = Path(__file__).resolve().parents[1]
    candidates = [
        engine_dir / "build" / "Release" / "cf_pipeline_v2.exe",
        engine_dir / "build" / "Debug" / "cf_pipeline_v2.exe",
        engine_dir / "build" / "cf_pipeline_v2.exe",
        engine_dir / "cf_pipeline_v2.exe",
        Path("cf_pipeline_v2.exe"),
        Path("cf_pipeline_v2"),
    ]
    for path in candidates:
        if path.exists():
            return str(path)
    return "cf_pipeline_v2"


def _run_once(cmd: list[str]) -> float:
    env = os.environ.copy()
    if env.get("CF_ALLOW_DIRECT_ENGINE", "").strip().lower() not in {"1", "true", "yes", "on"}:
        raise RuntimeError(
            "Direct engine benchmark is policy-gated. "
            "Set CF_ALLOW_DIRECT_ENGINE=1 explicitly to run this tool."
        )
    t0 = time.perf_counter()
    proc = subprocess.run(cmd, check=False, env=env)
    dt = time.perf_counter() - t0
    if proc.returncode != 0:
        raise RuntimeError(f"Command failed ({proc.returncode}): {' '.join(cmd)}")
    return dt


def _run_many(label: str, cmd: list[str], runs: int) -> None:
    samples: list[float] = []
    for _ in range(runs):
        samples.append(_run_once(cmd))
    med = statistics.median(samples)
    p95 = statistics.quantiles(samples, n=20)[-1] if len(samples) >= 2 else samples[0]
    print(f"{label}: median={med:.3f}s p95≈{p95:.3f}s runs={runs}")


def main() -> int:
    parser = argparse.ArgumentParser(description="Benchmark engine + step-internal parallelism controls.")
    parser.add_argument("--exe", default=_default_exe(), help="Path to cf_pipeline_v2 executable")
    parser.add_argument(
        "--pipeline",
        default=str(Path(__file__).resolve().parents[1] / "examples" / "omp_reduction_parallel.jsonld"),
        help="Pipeline JSON-LD (defaults to omp_reduction_parallel.jsonld)",
    )
    parser.add_argument("--runs", type=int, default=3, help="Number of runs per scenario")
    args = parser.parse_args()

    # The pipeline exercises cfbd:OmpReductionStep (cf_basic_dev). Ensure the plugin is built if needed.
    pipeline = str(Path(args.pipeline))

    base = [args.exe, "--pipeline", pipeline]

    try:
        _run_many("safe-default", base + ["--engine-threads", "2"], runs=args.runs)
        _run_many(
            "nested-allowed",
            base
            + [
                "--engine-threads",
                "2",
                "--max-total-threads",
                "16",
                "--default-step-threads",
                "8",
                "--respect-step-parallelism",
                "true",
            ],
            runs=args.runs,
        )
        _run_many(
            "budgeted",
            base
            + [
                "--engine-threads",
                "2",
                "--max-total-threads",
                "8",
                "--default-step-threads",
                "8",
                "--respect-step-parallelism",
                "true",
            ],
            runs=args.runs,
        )
        _run_many(
            "deterministic",
            base
            + [
                "--engine-threads",
                "2",
                "--max-total-threads",
                "8",
                "--default-step-threads",
                "8",
                "--deterministic",
            ],
            runs=args.runs,
        )
    except Exception as exc:
        print(f"bench_parallelism.py error: {exc}", file=sys.stderr)
        return 1

    return 0


if __name__ == "__main__":
    raise SystemExit(main())



from __future__ import annotations

import json
import os
import socket
import subprocess
import sys
import time
from pathlib import Path

import pytest

duckdb = pytest.importorskip("duckdb")


def _repo_root() -> Path:
    return Path(__file__).resolve().parents[4]


def _extend_sys_path() -> None:
    repo_root = _repo_root()
    src_paths = [
        repo_root / "sandcastle" / "cf_ontology" / "src",
        repo_root / "sandcastle" / "cf_opcua_server" / "src",
        repo_root / "sandcastle" / "cf_pipeline" / "cf_pipeline_engine",
        repo_root / "sandcastle" / "cf_datahive" / "src",
    ]
    for src_path in src_paths:
        src_path_str = str(src_path)
        if src_path.exists() and src_path_str not in sys.path:
            sys.path.insert(0, src_path_str)


def _find_engine_exe() -> Path | None:
    root = _repo_root()
    candidates = [
        root / "sandcastle" / "cf_pipeline" / "cf_pipeline_engine" / "build" / "Debug" / "cf_pipeline_v2.exe",
        root / "sandcastle" / "cf_pipeline" / "cf_pipeline_engine" / "build" / "Release" / "cf_pipeline_v2.exe",
        root / "sandcastle" / "cf_pipeline" / "cf_pipeline_engine" / "build" / "cf_pipeline_v2.exe",
        root / "sandcastle" / "cf_pipeline" / "cf_pipeline_engine" / "build" / "cf_pipeline_v2",
    ]
    for candidate in candidates:
        if candidate.exists():
            return candidate
    return None


def _plugins_available() -> bool:
    root = _repo_root()
    required = [
        root / "sandcastle" / "cf_basic_steps" / "cf_basic_io" / "src" / "cf_basic_io" / "bin" / "cf_basic_io_v2.dll",
        root / "sandcastle" / "cf_basic_steps" / "cf_basic_signal" / "src" / "cf_basic_signal" / "bin" / "cf_basic_signal_v2.dll",
        root / "sandcastle" / "cf_basic_steps" / "cf_basic_sinks" / "src" / "cf_basic_sinks" / "bin" / "cf_basic_sinks_v2.dll",
    ]
    return all(path.exists() for path in required)


def _pythonpath_env(base_env: dict[str, str]) -> dict[str, str]:
    root = _repo_root()
    additions = [
        root / "sandcastle" / "cf_ontology" / "src",
        root / "sandcastle" / "cf_opcua_server" / "src",
        root / "sandcastle" / "cf_pipeline" / "cf_pipeline_engine",
        root / "sandcastle" / "cf_datahive" / "src",
    ]
    current = base_env.get("PYTHONPATH", "")
    parts = [str(path) for path in additions]
    if current:
        parts.append(current)
    base_env["PYTHONPATH"] = os.pathsep.join(parts)
    return base_env


def _run_cmd(args: list[str], *, env: dict[str, str], cwd: Path) -> subprocess.CompletedProcess[str]:
    return subprocess.run(
        args,
        cwd=str(cwd),
        env=env,
        capture_output=True,
        text=True,
        check=False,
    )


def _wait_for_port(host: str, port: int, timeout_seconds: float) -> bool:
    deadline = time.monotonic() + timeout_seconds
    while time.monotonic() < deadline:
        try:
            with socket.create_connection((host, port), timeout=1.0):
                return True
        except OSError:
            time.sleep(0.2)
    return False


def test_opcua_fifo_avg_demo_persists_datahive_parquet(tmp_path: Path) -> None:
    _extend_sys_path()

    engine_exe = _find_engine_exe()
    if engine_exe is None:
        pytest.skip("cf_pipeline_v2 executable not found; build cf_pipeline_engine first")
    if not _plugins_available():
        pytest.skip("required step plugins not found; build cf_basic_io/cf_basic_signal/cf_basic_sinks first")

    from cf_datahive import DataHiveClient

    repo_root = _repo_root()
    pipeline_path = (
        repo_root
        / "sandcastle"
        / "cf_pipeline"
        / "cf_pipeline_engine"
        / "examples"
        / "opcua_fifo_avg_to_duckdb_parquet_triggered.jsonld"
    )
    semantics_dir = tmp_path / "semantics"
    semantics_dir.mkdir()
    workspace_root = tmp_path / "workspace"
    workspace_root.mkdir()

    env = _pythonpath_env(dict(os.environ))
    env["CF_PIPELINE_V2_EXE"] = str(engine_exe)

    server_proc: subprocess.Popen[str] | None = None

    try:
        cmd_base = [
            sys.executable,
            "-m",
            "cf_ontology",
        ]

        init_proc = _run_cmd(
            cmd_base + [
                "init",
                "--semantics-dir",
                str(semantics_dir),
                "--rebuild",
            ],
            env=env,
            cwd=tmp_path,
        )
        assert init_proc.returncode == 0, init_proc.stderr

        id_proc = _run_cmd(
            cmd_base + [
                "pipeline",
                "id",
                "--jsonld",
                str(pipeline_path),
            ],
            env=env,
            cwd=tmp_path,
        )
        assert id_proc.returncode == 0, id_proc.stderr
        pipeline_state_id = id_proc.stdout.strip().splitlines()[-1]
        assert pipeline_state_id

        commit_proc = _run_cmd(
            cmd_base + [
                "pipeline",
                "commit",
                "--semantics-dir",
                str(semantics_dir),
                "--pipeline-id",
                pipeline_state_id,
                "--jsonld",
                str(pipeline_path),
                "--user",
                "test",
                "--message",
                "integration",
            ],
            env=env,
            cwd=tmp_path,
        )
        assert commit_proc.returncode == 0, commit_proc.stderr

        activate_proc = _run_cmd(
            cmd_base + [
                "state",
                "activate",
                "--semantics-dir",
                str(semantics_dir),
                "--pipeline-id",
                pipeline_state_id,
                "--desired-state",
                "enabled",
            ],
            env=env,
            cwd=tmp_path,
        )
        assert activate_proc.returncode == 0, activate_proc.stderr

        server_proc = subprocess.Popen(
            [
                sys.executable,
                "-m",
                "cf_opcua_server.cli",
                "start",
                "--interval",
                "0.5",
                "--duration",
                "60",
            ],
            cwd=str(tmp_path),
            env=env,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            text=True,
        )
        assert _wait_for_port("127.0.0.1", 4840, timeout_seconds=15.0), "OPC UA server did not start"

        emit_event_proc = _run_cmd(
            cmd_base
            + [
                "state",
                "emit-event",
                "--semantics-dir",
                str(semantics_dir),
                "--pipeline-id",
                pipeline_state_id,
                "--event-type",
                "opcua_signal",
                "--source",
                "integration-test",
                "--dedupe-key",
                "integration-opcua-001",
            ],
            cwd=tmp_path,
            env=env,
        )
        assert emit_event_proc.returncode == 0, emit_event_proc.stderr

        run_loop_proc = _run_cmd(
            cmd_base
            + [
                "state",
                "run-loop",
                "--semantics-dir",
                str(semantics_dir),
                "--pipeline-id",
                pipeline_state_id,
                "--poll-interval",
                "0.2",
                "--once",
            ],
            cwd=tmp_path,
            env=env,
        )
        if run_loop_proc.returncode != 0 and "Failed to load plugin:" in run_loop_proc.stderr:
            pytest.skip("pipeline plugin load failed in this environment; rebuild plugins with matching toolchain")
        assert run_loop_proc.returncode == 0, run_loop_proc.stderr

        state_db = duckdb.connect(str(semantics_dir / "cf-pipeline-states.duckdb"), read_only=True)
        try:
            linkage_row = state_db.execute(
                """
                SELECT
                  re.event_id,
                  re.run_id,
                  re.status,
                  pr.triggered_by_event_id,
                  pr.status
                FROM run_events AS re
                JOIN pipeline_runs AS pr
                  ON pr.run_id = re.run_id
                WHERE re.pipeline_id = ? AND re.dedupe_key = ?
                ORDER BY re.created_at DESC
                LIMIT 1
                """,
                [pipeline_state_id, "integration-opcua-001"],
            ).fetchone()
        finally:
            state_db.close()
        assert linkage_row is not None
        assert str(linkage_row[0]) == str(linkage_row[3])
        assert linkage_row[1] is not None
        assert str(linkage_row[2]) == "consumed"
        assert str(linkage_row[4]) == "succeeded"

        client = DataHiveClient(str(workspace_root))
        runs = client.list_runs("opcua_fifo_avg")
        assert runs, "no datahive runs written for opcua_fifo_avg"

        run_id = runs[0].run_id
        run = client.open_run("opcua_fifo_avg", run_id)
        manifest_path = run.run_path / "manifest.json"
        assert manifest_path.exists()

        manifest = json.loads(manifest_path.read_text(encoding="utf-8"))
        assert manifest["status"] == "committed"
        assert isinstance(manifest.get("session_id"), str) and manifest["session_id"]
        assert isinstance(manifest.get("session_run_index"), int)
        assert f"run_id={run_id}" in str(manifest["tables"]["measurements"]["path"])
        assert run.run_path.parent.name == "runs"
        assert run.run_path.parent.parent.name == manifest["session_id"]

        table = client.read_table("opcua_fifo_avg", run_id, "measurements")
        payload = table.to_pydict()
        assert table.num_rows == 20
        assert payload["cycle_id"] == list(range(20))

        archive_candidates = list(workspace_root.rglob("archive.jsonl"))
        assert not archive_candidates
    finally:
        if server_proc is not None:
            server_proc.terminate()
            try:
                server_proc.wait(timeout=5)
            except subprocess.TimeoutExpired:
                server_proc.kill()
                server_proc.wait(timeout=5)



from __future__ import annotations

import argparse

from cf_pipeline_engine import cli


class _Proc:
    def __init__(self, returncode: int = 0) -> None:
        self.returncode = returncode


def _make_args() -> argparse.Namespace:
    return argparse.Namespace(
        pipeline="demo.jsonld",
        steps=None,
        plugins=None,
        engine_threads=0,
        max_total_threads=None,
        default_step_threads=None,
        respect_step_parallelism=None,
        deterministic=None,
        interval=0.0,
        iterations=1,
        duration=0.0,
        debug=False,
        exe="cf_pipeline_v2",
    )


def test_direct_engine_blocked_by_default(monkeypatch) -> None:
    monkeypatch.delenv("CF_ALLOW_DIRECT_ENGINE", raising=False)
    monkeypatch.setattr(cli, "_discover_plugin_dirs", lambda: [])

    def _should_not_run(*args, **kwargs):  # noqa: ANN001, ANN002
        raise AssertionError("subprocess.run must not be called when policy blocks direct start")

    monkeypatch.setattr(cli.subprocess, "run", _should_not_run)
    code = cli._run_from_args(_make_args())
    assert code == 2


def test_direct_engine_allowed_with_override(monkeypatch) -> None:
    monkeypatch.setenv("CF_ALLOW_DIRECT_ENGINE", "1")
    monkeypatch.setattr(cli, "_discover_plugin_dirs", lambda: [])
    seen: dict[str, object] = {}

    def _fake_run(cmd, check=False):  # noqa: ANN001, ANN202
        seen["cmd"] = cmd
        seen["check"] = check
        return _Proc(0)

    monkeypatch.setattr(cli.subprocess, "run", _fake_run)
    code = cli._run_from_args(_make_args())

    assert code == 0
    assert seen["check"] is False
    assert isinstance(seen["cmd"], list)
    assert "--pipeline" in seen["cmd"]


def test_direct_engine_not_allowed_with_invocation_mode_only(monkeypatch) -> None:
    monkeypatch.delenv("CF_ALLOW_DIRECT_ENGINE", raising=False)
    monkeypatch.setenv("CF_PIPELINE_INVOCATION_MODE", "state-run-loop")
    monkeypatch.setattr(cli, "_discover_plugin_dirs", lambda: [])

    def _should_not_run(*args, **kwargs):  # noqa: ANN001, ANN002
        raise AssertionError("subprocess.run must not be called when only invocation mode is set")

    monkeypatch.setattr(cli.subprocess, "run", _should_not_run)
    code = cli._run_from_args(_make_args())
    assert code == 2


def test_unified_cli_default_modules_excludes_engine_wrapper(monkeypatch) -> None:
    try:
        from cf_cli.cli import DEFAULT_MODULES
    except Exception:
        return
    assert "cf_pipeline_engine.cli" not in DEFAULT_MODULES



from __future__ import annotations

from importlib import metadata
from pathlib import Path
import os
import subprocess
import sys

import cf_pipeline_engine


def test_distribution_metadata_matches_module_version() -> None:
    assert metadata.version("cf-pipeline-engine") == cf_pipeline_engine.__version__


def test_packaged_native_assets_are_present() -> None:
    include_dir = Path(cf_pipeline_engine.cf_engine_include_path())
    assert include_dir.is_dir()
    assert (include_dir / "cf_step_abi.h").is_file()
    assert (include_dir / "cf_step_utils.h").is_file()
    assert (include_dir / "cf_plugin_table.h").is_file()

    assert Path(cf_pipeline_engine.cf_siggen_path()).is_file()
    assert Path(cf_pipeline_engine.cf_pipeline_v2_path()).is_file()
    assert Path(cf_pipeline_engine.cf_type_registry_path()).is_file()


def test_packaged_binaries_respond() -> None:
    siggen = Path(cf_pipeline_engine.cf_siggen_path())
    siggen_result = subprocess.run([str(siggen)], capture_output=True, text=True, check=False)
    siggen_output = f"{siggen_result.stdout}\n{siggen_result.stderr}"
    assert siggen_result.returncode != 0
    assert "Usage: cf_siggen" in siggen_output

    engine = Path(cf_pipeline_engine.cf_pipeline_v2_path())
    env = os.environ.copy()
    env["CF_ALLOW_DIRECT_ENGINE"] = "1"
    engine_result = subprocess.run([str(engine)], env=env, capture_output=True, text=True, check=False)
    engine_output = f"{engine_result.stdout}\n{engine_result.stderr}"
    assert engine_result.returncode == 2
    assert "Usage: cf_pipeline_v2" in engine_output


def test_resolve_executable_prefers_env_override(monkeypatch) -> None:
    monkeypatch.setenv("CF_PIPELINE_V2_EXE", "C:/custom/cf_pipeline_v2.exe")
    assert cf_pipeline_engine.resolve_cf_pipeline_v2_executable() == "C:/custom/cf_pipeline_v2.exe"


def test_cli_module_help_surface() -> None:
    cli_result = subprocess.run(
        [sys.executable, "-m", "cf_pipeline_engine.cli", "--help"],
        capture_output=True,
        text=True,
        check=False,
    )
    output = f"{cli_result.stdout}\n{cli_result.stderr}"
    assert cli_result.returncode == 0
    assert "Run a pipeline using the C++ v2 backend." in output



"""Python package wrapper for the C++ Cogniflow pipeline engine."""

from __future__ import annotations

from importlib import metadata
import os
from pathlib import Path
import shutil

__version__ = "0.1.4"


def _locate_distribution_file(*relative_candidates: str) -> Path:
    distribution = metadata.distribution("cf-pipeline-engine")
    files = distribution.files or []
    for candidate in relative_candidates:
        for entry in files:
            if entry.as_posix() == candidate:
                return Path(distribution.locate_file(entry)).resolve()
    raise FileNotFoundError(
        f"Could not locate any of {relative_candidates!r} in cf-pipeline-engine distribution files."
    )


def cf_pipeline_v2_path() -> str:
    return str(_locate_distribution_file("bin/cf_pipeline_v2.exe", "bin/cf_pipeline_v2"))


def cf_siggen_path() -> str:
    return str(_locate_distribution_file("bin/cf_siggen.exe", "bin/cf_siggen"))


def cf_engine_include_path() -> str:
    return str(_locate_distribution_file("include/cf_step_abi.h").parent)


def cf_type_registry_path() -> str:
    return str(_locate_distribution_file("bin/type_registry.v0.json"))


def _repo_build_engine_exe() -> str | None:
    engine_dirs = [
        Path("sandcastle/cf_pipeline/cf_pipeline_engine"),
        Path(__file__).resolve().parents[1],
    ]

    seen: set[str] = set()
    ordered_dirs: list[Path] = []
    for entry in engine_dirs:
        key = str(entry)
        if key in seen:
            continue
        seen.add(key)
        ordered_dirs.append(entry)

    config_dirs = ["Release", "RelWithDebInfo", "Debug", "MinSizeRel"]
    exe_names = ["cf_pipeline_v2.exe", "cf_pipeline_v2"]

    candidates: list[Path] = []
    for engine_dir in ordered_dirs:
        build_dir = engine_dir / "build"
        for exe_name in exe_names:
            candidates.append(build_dir / exe_name)
        for config_dir in config_dirs:
            for exe_name in exe_names:
                candidates.append(build_dir / config_dir / exe_name)
                candidates.append(build_dir / "bin" / config_dir / exe_name)

    for exe_name in exe_names:
        candidates.append(Path(exe_name))

    for candidate in candidates:
        if candidate.exists():
            return str(candidate)
    return None


def resolve_cf_pipeline_v2_executable() -> str:
    env_override = os.environ.get("CF_PIPELINE_V2_EXE")
    if env_override:
        return env_override

    try:
        return cf_pipeline_v2_path()
    except Exception:
        pass

    on_path = shutil.which("cf_pipeline_v2") or shutil.which("cf_pipeline_v2.exe")
    if on_path:
        return on_path

    repo_build = _repo_build_engine_exe()
    if repo_build:
        return repo_build

    return "cf_pipeline_v2"


__all__ = [
    "__version__",
    "cf_pipeline_v2_path",
    "cf_siggen_path",
    "cf_engine_include_path",
    "cf_type_registry_path",
    "resolve_cf_pipeline_v2_executable",
]
